;; -*- hen -*-
(declare (unit LWS-UA)
		 (uses LWS-headers
			   LWS-request
			   LWS-response)
		 (export <lws:UA>
				 initialize
				 request))

(use tinyclos
	 (srfi 13)
	 (srfi 18)
	 base64
	 regex
	 tcp)

(include "cleited-macros.scm")

(define-class <lws:UA> ()
  (agent           ;; libwww-scheme/#.##
   default-headers
   max-size
   max-redirect
   timeout))

(define-accessors <lws:UA>
  (agent
   default-headers
   max-size
   max-redirect
   timeout))


(define-method (initialize (this <lws:UA>) args)
  (call-next-method)

  (initialize-slots this (list 'agent           "Mozilla/5.0 (not really. I am libwww-scheme/0.1)"
							   'default-headers (make <lws:headers>)
							   'max-size        #f
							   'max-redirect    7
							   'timeout         180)))


(define-method (setup-request (this <lws:UA>) (req-source <lws:request>))
  (with-slots (agent default-headers) this

	  (define req (clone req-source))
	  
	  ;; headers 弄り
	  (let [(headers (lws:headers-of req))
			(uri     (lws:uri-of     req))
			(content (lws:content-of req))]

		;; userinfo があれば Authorization: に移す。
		(let [(userinfo (uri:userinfo-of uri))]
		  (when userinfo
			(set-header! headers "Authorization"
						 (sprintf "Basic ~a"
								  (base64:encode userinfo)))
			(uri:set-userinfo! uri #f)))
		
		;; Host: が無ければ uri のホスト部を Host: に移す。
		(unless (get-value headers "Host")
		  (set-header! headers "Host"
					   (let [(get-host (lambda ()
										 (if (string-index (uri:host-of uri) #\:)
											 (sprintf "[~a]" (uri:host-of uri)) ;; IPv6 addr
											 (uri:host-of uri))))]
						 (if (eq? #f (uri:port-of uri))
							 (get-host)
							 (sprintf "~a:~a" (get-host) (uri:port-of uri)))))
		  (uri:set-scheme!   uri #f)
		  (uri:set-host!     uri #f))

		;; Content-Length を設定
		(set-header! headers "Content-Length"
					 (->string (string-length content)))

		;; User-Agent を設定
		(init-header! headers "User-Agent" agent)

		;; default-headers を轉冩
		(map-header (lambda (key value)
					  (init-header! headers key value))
		    default-headers))

	  req))
	  

(define-method (try-request (this <lws:UA>) (req <lws:request>))
  (with-slots (timeout) this
			  
      (let* [(res       (make <lws:response>))
			 (io-thread (make-thread (lambda ()
									   (io-thread-main this req res))))]
		(condition-case
		 (begin
		   (thread-start! io-thread)

		   (if timeout
			   (thread-join! io-thread timeout)
			   (thread-join! io-thread))

		   res)

		 [exn ()
		    (cond ((join-timeout-exception? exn)
				   ;; タイムアウトした
				   (make <lws:response>
					 code:    408
					 message: "Request Timeout"))

				  (else
				   (make <lws:response>
					 code:    500
					 message: (let* [(acc (condition-property-accessor 'exn 'message))
									 (msg (acc exn))]
								(->string msg)))))]))))


(define-method (io-thread-main (this <lws:UA>) (req <lws:request>) (res <lws:response>))
  
  (let* [(host     (get-value (lws:headers-of req) "Host"))
		 (uri      (make <uri> (sprintf "http://~a/" host)))
		 (hostname (uri:host-of uri))
		 (port     (or (uri:port-of uri) (uri:default-port uri)))]

	(let-values [((in out) (tcp-connect hostname port))]

	   (dynamic-wind
		(lambda ()
		  #f)

		(lambda ()
		  (fprintf out "~a ~a HTTP/1.0\r\n"
				   (lws:method-of req)
				   (uri->string (lws:uri-of req)))

		  (map-header (lambda (key value)
						(fprintf out "~a: ~a\r\n" key value))
					  (lws:headers-of req))

		  (fprintf out "\r\n")
		  (fprintf out "~a" (lws:content-of req))
		  (flush-output out)

		  ;; ここでリクエスト終わり。應答を讀み始める。
		  (let* [(resp-line (read-line in 1024))
				 (m         (string-search (regexp "^HTTP/\\d+\\.\\d+ (\\d+) (.+)$")
										   resp-line))]
			(lws:set-code!    res (string->number (cadr m)))
			(lws:set-message! res (caddr m)))

		  (let loop [(line (read-line in 1024))]
			(when (not (equal? line ""))
			  
			  (let [(m (string-search (regexp "^([^: ]+) *: *(.+)$")
									  line))]
				(push-header! (lws:headers-of res)
							  (cadr  m)
							  (caddr m)))
			  
			  (loop (read-line in 1024))))
		  
		  ;; ここで應答ヘッダ終わり。次は content を讀む。
		  ;; FIXME: Content-Length を見る
		  ;; FIXME: max-size の考慮
		  (lws:set-content! res (read-string #f in)))

		(lambda ()
		  (close-input-port  in)
		  (close-output-port out))))))


(define-method (request-and-redirect (this <lws:UA>)
									 (req-source <lws:request>)
									 redir-remain)
  (with-slots (max-redirect) this
      (let* [(req      (setup-request this req-source))
			 (res      (try-request   this req       ))
			 (location (get-value (lws:headers-of res) "Location"))]

		(cond ((and (lws:redirect? res) location)
			   ;; リダイレクトが指示された場合は、可能な範圍で再リクエストする。
			   (if (or (not redir-remain)
					   (< 0 redir-remain))
					 
				   (let [(new-req (clone req-source))]
						 
					 (lws:set-uri! new-req (make <uri> location))
					 (request-and-redirect this new-req
										   (if redir-remain
											   (- redir-remain 1)
											   #f)))

				   (make <lws:response>
					 code:    500
					 message: "Too many redirects")))

			  (else res)))))


(define-method (request (this <lws:UA>) (req <lws:request>))
  (with-slots (max-size max-redirect) this

	  (request-and-redirect this req max-redirect)))

							   