;; -*- hen -*-
;; RSSA - RSS Abstraction
;;
(declare (unit RSSA)
         (export <rssa:feed>
                 <rssa:item>
                 initialize

                 rssa:version-of
                 rssa:channel-of
                 rssa:items-of

                 rssa:title-of
                 rssa:link-of
                 rssa:date-of
                 rssa:description-of))

(use tinyclos
     (srfi 1)
     ssax)

(include "cleited-macros.scm")

(define rssa:namespace-prefixes
  '((rdf     . "http://www.w3.org/1999/02/22-rdf-syntax-ns#" )
    (dc      . "http://purl.org/dc/elements/1.1/"            )
    (dcterms . "http://purl.org/dc/terms/"                   )
    (sy      . "http://purl.org/rss/1.0/modules/syndication/")
    (admin   . "http://webns.net/mvcb/"                      )
    (rss     . "http://purl.org/rss/1.0/"                    )
    (rss091  . "http://purl.org/rss/1.0/modules/rss091#"     )
    (rdf     . "http://my.netscape.com/rdf/simple/0.9/"      )))

(define (rssa:error msg . args)
  (signal (make-composite-condition
           (make-property-condition 'rssa)
           (make-property-condition 'exn 'message msg 'arguments args))))

(define (rssa:warn fstr . args)
  (fprintf (current-error-port) "RSSA Warning: ~?~%" fstr args))

;; -[ rssa:feed ]---------------------------------------------------------------
;;
(define-class <rssa:feed> ()
  (version   ; <string>
   channel   ; <rssa:item>
   items))   ; <rssa:item>

(define-accessors <rssa:feed>
  (version
   channel
   items)
  prefix: rssa:)

(define-method (initialize (this <rssa:feed>) args)
  (call-next-method)

  (define (init #!key in-port)

    (assert (or (null?       in-port)
                (input-port? in-port)))

    (initialize-slots this
                      (list 'version "1.0"
                            'channel #f
                            'items   (list)))

    (when in-port
      (rssa:parse! this in-port)))

  (apply init args))


(define rssa:cleanup
  (match-lambda [('*TOP* items ...)
                 (let loop ([items items])
                   (match (car items)
                          [('*PI*         . _) (loop (cdr items))]
                          [('*NAMESPACES* . _) (loop (cdr items))]
                          [x x]))]
                [x x]))
                          

(define-method (rssa:parse! (this <rssa:feed>)
                            (port <input-port>))
  (let ([sxml (SSAX:XML->SXML port rssa:namespace-prefixes)])

    (match (rssa:cleanup sxml)
           [('rss ('@ attrs ...) data ...)
            (rssa:find-version! this attrs)
            (for-each (cut rssa:traverse-element! this <>)
                      data)]

           [('rdf:RDF data ...)
            (rssa:set-version! this "1.0")
            (for-each (cut rssa:traverse-element! this <>)
                      data)]

           [_
            (rssa:error "invalid root element" sxml)])))

(define-method (rssa:find-version! (this <rssa:feed>) attrs)
  (for-each (match-lambda [('version v)
                           (rssa:set-version! this v)])
            attrs))

(define-method (rssa:traverse-element! (this <rssa:feed>) sxml #!optional item)
  (match sxml
         [((or 'channel 'rss:channel 'rdf:channel) elts ...)
          (let [(channel (make <rssa:item>))]
            (rssa:set-channel! this channel)
            (for-each (cut rssa:traverse-element! this <> item)
                      elts))]

         [((or 'item 'rss:item) elts ...)
          (let [(item (make <rssa:item>))]
            (rssa:append-item! this item)
            (for-each (cut rssa:traverse-element! this <> item)
                      elts))]

         [((or 'title 'rss:title 'dc:title) title)
          (let [(target (or item (rssa:channel-of this)))]
            (if target
                (rssa:set-title! target title)
                (rssa:warn "tag `title' with content ~S in wrong context" title)))]

         [((or 'link 'rss:link) link)
          (let [(target (or item (rssa:channel-of this)))]
            (if target
                (rssa:set-link! target link)
                (rssa:warn "tag `link' with content ~S in wrong context" link)))]

         [((or 'dcterms:modified 'dc:date 'pubDate 'rss091:pubDate) date)
          (let [(target (or item (rssa:channel-of this)))]
            (if target
                (rssa:set-date! target date)
                (rssa:warn "tag `date' with content ~S in wrong context" date)))]

         [((or 'description 'rss:description 'dc:description 'dcterms:description) description)
          (let [(target (or item (rssa:channel-of this)))]
            (if target
                (rssa:set-description! target description)
                (rssa:warn "tag `description' with content ~S in wrong context" description)))]

         [(tag ('@ . _) . more)
          (rssa:traverse-element! this (cons tag more) item)]

         [(tag . more)
          (for-each (cut rssa:traverse-element! this <> item)
                    more)]

         [_ #f]))

;; -[ rssa:item ]---------------------------------------------------------------
(define-class <rssa:item> ()
  (title         ; <string>
   link          ; <string>
   date          ; <string>
   description)) ; <string>

(define-accessors <rssa:item>
  (title
   link
   date
   description)
  prefix: rssa:)


(define-method (initialize (this <rssa:item>) args)
  (call-next-method)

  (initialize-slots this
                    (list 'title       #f
                          'link        #f
                          'date        #f
                          'description #f)))


(define-method (rssa:append-item! (feed <rssa:feed>)
                                  (item <rssa:item>))
  (rssa:set-items! feed
                   (append (rssa:items-of feed)
                           (list item))))
