;; -*- hen -*-
(declare (unit account)
		 (uses cleited-passive-node)
		 (export account:install-methods!
				 account:check))

(use tinyclos
	 sqlite3
	 sha2
	 base64)


(define (account:install-methods! node)
  (assert (instance-of? node <cleited-passive-node>))
  
  (wrap-method! node 'account:check  account:check)
  (wrap-method! node 'account:get    account:get)
  (wrap-method! node 'account:list   account:list)
  (wrap-method! node 'account:add    account:add)
  (wrap-method! node 'account:delete account:delete)
  (wrap-method! node 'account:update account:update))


(define (account:check arg db-pod #!key required-uid need-root)
  (assert (pair? arg))
  (assert (instance-of? db-pod <db-pod>))
   ;; required-uid: 等しい事が要求されるユーザ ID。今囘認證されたユーザ
   ;;   の ID がこれに一致してゐる時は、need-root が #t であったとして
   ;;   も操作が許可される。
  (let ((user-id (get-arg arg 'user-id check: string?))
		(passwd  (get-arg arg 'passwd  check: string?)))

	(let* ((acc #f)
		   (proc (lambda (user-id name is-root)
				   (set! acc `((user-id . ,user-id)
							   (name    . ,name   )
							   (is-root . ,(number->true-or-false is-root)))))))
	  (db-pod:run db-pod
	      (lambda (db)
			(sqlite3:for-each-row
			 proc db (string-append "SELECT user_id, name, is_root    "
									"  FROM accounts                  "
									" WHERE user_id = ? AND passwd = ?")
			 user-id
			 (base64:encode (sha256:binary-digest passwd)))))

	  (unless acc
	    (error "The user ID or password is invalid."))
	  
	  (if need-root
		  (if (and required-uid
				   (equal? user-id required-uid))
			  acc

			  (or (eq? 'true (cdr (assq 'is-root acc)))
				  (error "Operation not permitted"))))
	  
	  acc)))


(define (account:get arg db-pod)
  (let [(auth    (get-arg arg 'auth    check: list?))
        (user-id (get-arg arg 'user-id check: string?))]

    (account:check auth db-pod required-uid: user-id need-root: #t)

    (db-pod:run db-pod
        (lambda (db)
          (condition-case
           (let [(row (sqlite3:first-row
                       db (string-append "SELECT user_id, name, is_root"
                                         "  FROM accounts              "
                                         " WHERE user_id = ?           ")
                       user-id))]
             `((user-id . ,(car row))
               (name    . ,(cadr row))
               (is-root . ,(number->true-or-false (caddr row)))))
                  
           [(exn sqlite3)
            (error (sprintf "No such user `~a'" user-id))])))))


(define (account:list arg db-pod)
  (let [(auth (get-arg arg 'auth check: list?))]
    
    (account:check auth db-pod need-root: #t)
   
    (db-pod:run db-pod
        (lambda (db)
          (sqlite3:map-row (lambda (user-id name is-root)
                             `((user-id . ,user-id)
                               (name    . ,name   )
                               (is-root . ,(number->true-or-false is-root))))
                           db
                           "SELECT user_id, name, is_root FROM accounts")))))


(define (account:add arg db-pod)
  (let ((auth    (get-arg arg 'auth    check: list?  ))
		(user-id (get-arg arg 'user-id check: string?))
		(name    (get-arg arg 'name    check: string?))
		(passwd  (get-arg arg 'passwd  check: string?))
		(is-root (get-arg arg 'is-root check: true-or-false?)))

	(account:check auth db-pod need-root: #t)

	(db-pod:run db-pod
	    (lambda (db)
		  (sqlite3:exec db "BEGIN IMMEDIATE")

		  (unless (= 0
					 (sqlite3:first-result
					  db "SELECT COUNT(*) FROM accounts WHERE user_id = ?"
					  user-id))
	 
			(error (format "User ~s is already there" user-id)))

		  (sqlite3:exec db 
						(string-append "INSERT INTO accounts"
									   "       (user_id, passwd, name, is_root)"
									   "VALUES (      ?,      ?,    ?,       ?)")
						user-id
						(base64:encode (sha256:binary-digest passwd))
						name
						(true-or-false->number is-root))

		  (sqlite3:exec db "COMMIT")))
	
	'ok))

(define (account:delete arg db-pod)
  (let ((auth    (get-arg arg 'auth    check: list?))
		(user-id (get-arg arg 'user-id check: string?)))

	(account:check auth db-pod required-uid: user-id need-root: #t)

	(db-pod:run db-pod
	    (lambda (db)
		  (sqlite3:exec
		   db "DELETE FROM accounts WHERE user_id = ?" user-id)))
	
	'ok))

(define (account:update arg db-pod)
  (let ((auth    (get-arg arg 'auth    check: list?                      ))
		(user-id (get-arg arg 'user-id check: string?                    ))
		(passwd  (get-arg arg 'passwd  check: string?        optional: #t))
		(name    (get-arg arg 'name    check: string?        optional: #t))
		(is-root (get-arg arg 'is-root check: true-or-false? optional: #t)))

	(let ((acc (account:check auth db-pod required-uid: user-id need-root: #t)))

	  ;; root でない者が is-root を變更する事は出來ない。
	  (and is-root
		   (eq? 'true  is-root)
		   (eq? 'false (cdr (assq 'is-root acc)))
		   (error
			"You can't make yourself a root hence you are not a root."))

	  (db-pod:run db-pod
	      (lambda (db)
			(sqlite3:exec db "BEGIN")

			(if passwd
				(sqlite3:exec db
							  "UPDATE accounts SET passwd = ? WHERE user_id = ?"
							  (base64:encode (sha256:binary-digest passwd))
							  user-id))
			
			(if name
				(sqlite3:exec db
							  "UPDATE accounts SET name = ? WHERE user_id = ?"
							  name user-id))

			(if is-root
				(sqlite3:exec db
							  "UPDATE accounts SET is_root = ? WHERE user_id = ?"
							  (true-or-false->number is-root) user-id))

			(sqlite3:exec db "COMMIT")))

	  'ok)))
