;; -*- hen -*-
(declare (unit db-setup)
		 (export db-setup))

(use sqlite3
     sha2
     base64)

(define *config-defaults*
  `( ((key         . max-items-per-feed)
      (value       . 999               )
      (type        . natural           )
      (description
       . ,(string-append
           "Limit of the number of feed items which one feed can have. "
           "When this limit is exceeded, older feed items will be deleted.")))
     
     ((key         . polling-interval-minute)
      (value       . 180                    )
      (type        . natural                )
      (description
       . ,(string-append
           "Interval of the feed polling in minute. You shouldn't set "
           "this value too small not to overload the net."))) ))

(define (db-setup db)
  ;; config
  (sqlite3:exec
   db (string-append "CREATE TABLE IF NOT EXISTS config (     "
					 "  conf_key    TEXT NOT NULL PRIMARY KEY,"
					 "  conf_val    TEXT NOT NULL,            "
                     "  conf_type   TEXT NOT NULL,            "
                     "  description TEXT NOT NULL             "
					 ");                                      "))
  
  ;; config デフォルト設定
  (for-each (lambda (entry)
			  (define key  (format "~s" (cdr (assq 'key   entry))))
			  (define val  (format "~s" (cdr (assq 'value entry))))
              (define type (format "~s" (cdr (assq 'type  entry))))
              (define desc (cdr (assq 'description entry)))
			  (if (= 0
					 (sqlite3:first-result 
					  db (string-append
						  "SELECT COUNT(*)    "
						  "  FROM config      "
						  " WHERE conf_key = ?") key))
				  (sqlite3:exec
				   db (string-append
                       "REPLACE INTO config                                 "
                       "        (conf_key, conf_val, conf_type, description)"
                       " VALUES (       ?,        ?,         ?,           ?)")
                   key val type desc)))
			*config-defaults*)

  ;; accounts
  (sqlite3:exec
   db (string-append "CREATE TABLE IF NOT EXISTS accounts ("
					 "  user_id  TEXT NOT NULL PRIMARY KEY,"
					 "  passwd   TEXT NOT NULL,            "
					 "  name     TEXT NOT NULL,            "
					 "  is_root  INTEGER NOT NULL          "
					 ")                                    "))
  (condition-case
   (sqlite3:exec
	   db (string-append "CREATE TRIGGER account_cleanup   "
						 "  DELETE ON accounts             "
						 "  FOR EACH ROW                   "
						 "  BEGIN                          "
						 
						 "    DELETE FROM subscription     "
						 "     WHERE user_id = OLD.user_id;"
						 
						 "    DELETE FROM read_items       "
						 "     WHERE user_id = OLD.user_id;"
						 
						 "  END                            "))
   [(exn sqlite3) #f])

  ;; root アカウント
  (if (= 0
	     (sqlite3:first-result
		  db "SELECT COUNT(*) FROM accounts WHERE is_root != 0"))
	  (sqlite3:exec
	   db (string-append "INSERT INTO accounts                     "
						 "       (user_id, passwd,   name, is_root)"
						 "VALUES ( 'root',      ?, 'Root',       1)")
	   (base64:encode (sha256:binary-digest ""))))

  ;; feeds
  (sqlite3:exec
   db (string-append "CREATE TABLE IF NOT EXISTS feeds (             "
					 "  feed_id     INTEGER NOT NULL                 "
					 "                PRIMARY KEY AUTOINCREMENT,     "
					 "  feed_uri    TEXT NOT NULL,                   "
                     "  site_uri    TEXT NOT NULL,                   "
					 "  lastmod     INTEGER NOT NULL,                "
					 "  title       TEXT NOT NULL,                   "
					 "  description TEXT NOT NULL,                   "
					 "  status      TEXT NOT NULL,                   "
					 "    /* 'ok'              : no error            "
					 "     * 'transport-error' : e.g. 404 Not Found  "
					 "     * 'feed-error'      : e.g. XML parse error"
					 "     */                                        "
					 "  lastcheck   INTEGER NOT NULL,                "
					 "                                               "
					 "  UNIQUE (feed_uri)                            "
					 ");                                             "))
  (condition-case
   (sqlite3:exec
       db (string-append "CREATE TRIGGER feed_cleanup      "
						 "  DELETE ON feeds                "
						 "  FOR EACH ROW                   "
						 "  BEGIN                          "

						 "    DELETE FROM items            "
						 "     WHERE feed_id = OLD.feed_id;"
						 
						 "    DELETE FROM subscription     "
						 "     WHERE feed_id = OLD.feed_id;"

						 "  END                            "))
   [(exn sqlite3) #f])

  ;; feed_source -- RSSソース (パースエラー發生時に原因を探る爲)
  (sqlite3:exec
   db (string-append "CREATE TABLE IF NOT EXISTS feed_source ("
                     "  feed_uri TEXT NOT NULL,               "
                     "  source   TEXT NOT NULL,               "
                     "                                        "
                     "  PRIMARY KEY (feed_uri)                "
                     ");                                      "))

  ;; items -- 各フィード内の項目
  (sqlite3:exec
   db (string-append
	   ;; 各フィード内の最大項目數は、config 表で制限を掛ける。項
	   ;; 目數が最大値を越えたら、古いものから削除される。
	   "CREATE TABLE IF NOT EXISTS items (        "
	   "  item_id     INTEGER NOT NULL            "
	   "                PRIMARY KEY AUTOINCREMENT,"
	   "  feed_id     INTEGER NOT NULL,           "
	   "  item_uri    TEXT NOT NULL,              "
	   "  title       TEXT NOT NULL,              "
	   "  description TEXT NOT NULL,              "
	   "  date        INTEGER NOT NULL,           "
	   "                                          "
	   "  UNIQUE (feed_id, item_uri)              "
	   ");                                        "))
  (condition-case
   (sqlite3:exec
       db (string-append "CREATE TRIGGER item_cleanup      "
						 "  DELETE ON items                "
						 "  FOR EACH ROW                   "
						 "  BEGIN                          "

						 "    DELETE FROM read_items       "
						 "     WHERE item_id = OLD.item_id;"

						 "  END                            "))
   [(exn sqlite3) #f])

  ;; subscription -- 各員の登録してゐるフィード
  (sqlite3:exec
   db (string-append
	   ;; ランクは付けた本人のみが見る事の出來るものである。他のユーザーの
	   ;; ランク付けによって計算された人氣度をユーザーに宣傳する爲に用ゐて
	   ;; はならない。これはデザイン上の決定である。
	   "CREATE TABLE IF NOT EXISTS subscription ("
	   "  feed_id  INTEGER NOT NULL,             "
	   "  user_id  TEXT NOT NULL,                "
	   "  rank     INTEGER NOT NULL DEFAULT 3,   "
	   "           /* 1 (min) to 5 (max) */      "
	   "                                         "
	   "  PRIMARY KEY (feed_id, user_id)         "
	   ");                                       "))

  ;; read_items -- 各員の讀んだ項目
  (sqlite3:exec
   db (string-append "CREATE TABLE IF NOT EXISTS read_items ("
					 "  item_id INTEGER NOT NULL,            "
					 "  user_id TEXT NOT NULL,               "
					 "                                       "
					 "  PRIMARY KEY (item_id, user_id)       "
					 ");                                     ")))
