;; -*- hen -*-
(declare (unit feed)
		 (uses cleited-passive-node
			   cleited-utils
			   uri
			   account
			   LWS-headers
			   LWS-request
			   LWS-response
			   LWS-UA
               logger
               RSSA)
		 (export feed:install-methods!
				 feed:check!))

(use (srfi 1)
	 tinyclos
	 sqlite3
	 rfc3339
	 regex-case
	 sxml-tools
	 htmlprag)


(define (feed:install-methods! node)
  (assert (instance-of? node <cleited-passive-node>))

  (wrap-method! node 'feed:subscribe   feed:subscribe  )
  (wrap-method! node 'feed:discover    feed:discover   )
  (wrap-method! node 'feed:list        feed:list       )
  (wrap-method! node 'feed:unsubscribe feed:unsubscribe)
  (wrap-method! node 'feed:get-info    feed:get-info   )
  (wrap-method! node 'feed:set-rank    feed:set-rank   ))


(define (feed:find-feed-id-by-uri db-pod uri)
  (set! uri (util:clean-uri uri))

  (db-pod:run db-pod
      (lambda (db)
		(condition-case
		 (sqlite3:first-result db
							   "SELECT feed_id FROM feeds WHERE feed_uri = ?"
							   (uri->string uri))
		 [(exn sqlite3) #f]))))


(define (feed:get-last-modified db-pod feed-id)
  ;; feed-id は #f でも良く、その場合は常に #f が返る。
  (if feed-id
	  (db-pod:run db-pod
	      (lambda (db)
			(condition-case (sqlite3:first-result
							 db "SELECT lastmod FROM feeds WHERE feed_id = ?"
							 feed-id)
							[(exn sqlite3) #f])))

	  #f))


(define (update-status! db-pod feed-id status)
  (db-pod:run db-pod
      (lambda (db)
		(sqlite3:update db (string-append "UPDATE feeds         "
										  "   SET status    = ?,"
										  "       lastcheck = ? "
										  " WHERE feed_id   = ? ")
						(symbol->string status)
						(current-seconds)
						feed-id))))


(define (feed:check! uri db-pod)
  (set! uri (util:clean-uri uri))

  (let* [(feed-id (feed:find-feed-id-by-uri db-pod uri))
		 (lastmod (feed:get-last-modified db-pod feed-id))
		 (ua      (make <lws:UA>))
		 (req     (make <lws:request> 'GET uri))]

	(if lastmod
		(set-header! (lws:headers-of req)
					 "If-Modified-Since" (util:epoch->gmt-string lastmod)))

	(let [(res (request ua req))]
	  ;; 200 の場合は RSS をパース。
	  ;; 304 Not Modified の場合は何もせずに終了。
	  ;; それ以外の場合は轉送エラーと見做し、feeds を更新してエラー終了。
	  (cond [(= 200 (lws:code-of res))
			 (parse-and-store-rss! db-pod uri (lws:content-of res))]

			[(= 304 (lws:code-of res))
			 (update-status! db-pod feed-id 'ok)]

			[else
			 (if feed-id
				 (update-status! db-pod feed-id 'transport-error))
			 (error `(transport-error . ,(lws:status-of res)))]))))


(define (parse-and-store-rss! db-pod uri rss-string)

  ;; RSS を保存 (パースエラー時の原因究明の爲)
  (define (save-feed-text utf8-feed-str)
    (db-pod:run db-pod
        (lambda (db)
          (sqlite3:exec db (string-append "REPLACE INTO feed_source  "
                                          "        (feed_uri, source)"
                                          " VALUES (?       , ?     )")
                        (uri->string uri)
                        utf8-feed-str))))

  (define-macro (wrap-feed-error feed-id . body)
    `(condition-case
      (begin ,@body)
      [exn ()
         (if ,feed-id
             (update-status! db-pod ,feed-id 'feed-error))
         (error `(feed-error . ,((condition-property-accessor 'exn 'message) exn)))]))

  ;; RSS をパース
  ;; パースエラーの場合は feeds を更新して終了。
  (let* [(feed-id       (feed:find-feed-id-by-uri db-pod uri))
         (utf8-feed-str (wrap-feed-error feed-id
                             (let [(result (util:xml->utf8-xml rss-string))]
                               (save-feed-text result)
                               result)))
         (feed          (wrap-feed-error feed-id
                             (make <rssa:feed> in-port: (open-input-string utf8-feed-str))))]

	;; ここに來たと云ふ事は、RSSのパースに成功した。
	(db-pod:run db-pod
	    (lambda (db)
		  (if feed-id
			  ;; 更新の場合
              (begin
                (sqlite3:exec db (string-append "UPDATE feeds           "
                                                "   SET title       = ?,"
                                                "       site_uri    = ?,"
                                                "       description = ?,"
                                                "       status      = ?,"
                                                "       lastcheck   = ? "
                                                " WHERE feed_id     = ? ")
                              (or (rssa:title-of       (rssa:channel-of feed)) "")
                              (or (rssa:link-of        (rssa:channel-of feed)) "")
                              (or (rssa:description-of (rssa:channel-of feed)) "")
                              "ok"
                              (current-seconds)
                              feed-id)
                ;(logger 'parse-and-store-rss! 'info
                ;        "updated feed metadata: feed_id=[~a] title=[~a] feed_uri=[~a]"
                ;        feed-id
                ;        (or (rssa:title-of       (rssa:channel-of feed)) "")
                ;        (uri->string uri)))
                )

			  ;; 新規追加の場合
              (begin
                (sqlite3:exec
                 db (string-append "REPLACE INTO feeds                                                          "
                                   "        (feed_uri, lastmod, site_uri, title, description, status, lastcheck)"
                                   " VALUES (       ?,       ?,        ?,     ?,           ?,      ?,         ?)")
                 (uri->string uri)
                 (current-seconds)
                 (or (rssa:link-of        (rssa:channel-of feed)) "")
                 (or (rssa:title-of       (rssa:channel-of feed)) "")
                 (or (rssa:description-of (rssa:channel-of feed)) "")
                 "ok"
                 (current-seconds))
                (logger 'parse-and-store-rss! 'info
                        "inserted feed metadata: title=[~a] feed_uri=[~a]"
                        (or (rssa:title-of       (rssa:channel-of feed)) "")
                        (uri->string uri))))))

    (set! feed-id (feed:find-feed-id-by-uri db-pod uri))

	(db-pod:run db-pod
	    (lambda (db)
		  ;; トランザクション内で items テーブルを更新する。
		  (sqlite3:exec db "BEGIN IMMEDIATE")

		  (define feed-updated #f)
		  (define (install-item item)
			(let [(item-uri (uri->string (uri:canonicalize (make <uri>
															 (rssa:link-of item)))))
				  (date     (string->rfc3339-record (or (rssa:date-of item) "")))]

			  ;; items 内に古いデータがあれば更新。無ければ追加。更新の場合は、
			  ;; 更新時刻または内容が變化した場合は未讀にする。
			  (if (= 0 (sqlite3:first-result db
					       (string-append "SELECT COUNT(*)                    "
										  "  FROM items                       "
										  " WHERE feed_id = ? AND item_uri = ?")
						   feed-id item-uri))
				  ;; 新規追加
				  (begin
					(sqlite3:exec db
					    (string-append "INSERT INTO items                              "
									   "       (feed_id, item_uri, title, description, date)"
									   "VALUES (      ?,        ?,     ?,           ?,    ?)")
						feed-id
						item-uri
						(or (rssa:title-of       item) "")
						(or (rssa:description-of item) "")
						(if date
							(util:rfc3339->epoch date)
							(current-seconds)))
                    (logger 'parse-and-store-rss! 'info
                            "inserting feed item: feed_id=[~a] feed_uri=[~a] item_uri=[~a] title=[~a]"
                            feed-id
                            (uri->string uri)
                            item-uri
                            (or (rssa:title-of       item) ""))
					(set! feed-updated #t))

				  ;; 更新
				  (let* [(old (sqlite3:first-row db
							      (string-append "SELECT item_id, title, description, date"
												 "  FROM items                            "
												 " WHERE feed_id = ? AND item_uri = ?     ")
								  feed-id item-uri))
						 (old-item-id (car    old))
						 (old-title   (cadr   old))
						 (old-desc    (caddr  old))
						 (old-date    (cadddr old))]
		 
					(if (or (not (equal? old-title (or (rssa:title-of       item) "")))
							(not (equal? old-desc  (or (rssa:description-of item) ""))))
						(begin
						  (sqlite3:exec db
						      (string-append "UPDATE items                               "
											 "   SET title = ?, description = ?, date = ?"
											 " WHERE item_id = ?                         ")
							  (or (rssa:title-of       item) "")
							  (or (rssa:description-of item) "")
							  (if date
								  (util:rfc3339->epoch date)
								  (current-seconds))
							  old-item-id)
				 
						  (if (or (not (equal? old-desc (or (rssa:description-of item) "")))
								  (and date
									   (not (equal? old-date date))))
							  ;; read_items から削除するだけで未讀になる。
							  (sqlite3:exec db (string-append "DELETE FROM read_items"
															  " WHERE item_id = ?    ")
											old-item-id))

                          (logger 'parse-and-store-rss! 'info
                            "updating feed item: feed_id=[~a] feed_uri=[~a] item_id=[~a] item_uri=[~a] title=[~a]"
                            feed-id
                            (uri->string uri)
                            old-item-id
                            item-uri
                            (or (rssa:title-of       item) ""))
						  
						  (set! feed-updated #t)))))))
				 
		  ;; 各 feed item を格納。
		  (for-each install-item
					(rssa:items-of feed))

		  ;; feed item の個數が制限を越えたら、古いものから削除する。
		  (let [(item-limit (string->number (util:get-config db 'max-items-per-feed)))
				(num-items  (sqlite3:first-result db (string-append "SELECT COUNT(*)   "
																	"  FROM items      "
																	" WHERE feed_id = ?")
												  feed-id))]
			(sqlite3:map-row (lambda (item-id item-uri title)
							   (sqlite3:exec db (string-append "DELETE FROM items "
															   " WHERE item_id = ?")
											 item-id)
                               (logger 'parse-and-store-rss! 'info
                                       (string-append "deleting feed item due to overflow: "
                                                      "feed_id=[~a] feed_uri=[~a] item_id=[~a] "
                                                      "item_uri=[~a] title=[~a]")
                                       feed-id
                                       (uri->string uri)
                                       item-id
                                       item-uri
                                       title))
							 db
							 (string-append "SELECT item_id, item_uri, title"
											"  FROM items                   "
											" WHERE feed_id = ?             "
											" ORDER BY date ASC             "
											" LIMIT ?                       ")
							 feed-id
							 (if (< item-limit num-items)
								 (- num-items item-limit)
								 0)))
	
		  ;; 更新された feed item が存在するなら、feeds.lastmod を更新する。
		  (if feed-updated
			  (sqlite3:exec db (string-append "UPDATE feeds      "
											  "   SET lastmod = ?"
											  " WHERE feed_id = ?")
							(current-seconds)
							feed-id))

		  (sqlite3:exec db "COMMIT")))))


(define (feed:subscribe arg db-pod)
  (let ((auth (get-arg arg 'auth     check: list?))
		(uri  (get-arg arg 'feed-uri checK: string?)))

	(account:check auth db-pod)

	;; 今囘登録しようとしてゐる feed uri が既に feeds テーブル内に存在
	;; するのであれば、その uri が本當に feed であるのかどうかを確かめ
	;; る必要が無い。
	(define feed-id (feed:find-feed-id-by-uri db-pod uri))

	(unless feed-id
	  (feed:check! uri db-pod)
      (set! feed-id (feed:find-feed-id-by-uri db-pod uri)))

	(db-pod:run db-pod
	    (lambda (db)
		  (sqlite3:exec db (string-append "INSERT OR IGNORE INTO subscription"
										  "       (feed_id, user_id)         "
										  "VALUES (?      , ?      )         ")
						feed-id
						(get-arg auth 'user-id))))
	
	'ok))


(define (feed:discover arg db-pod)
  (let [(uri (get-arg arg 'uri check: string?))]

	(set! uri (util:clean-uri uri))

	(let* [(ua  (make <lws:UA>))
		   (req (make <lws:request> 'GET uri))
		   (res (request ua req))]
	  ;; 200 の場合は HTML をパース。
	  ;; それ以外の場合は轉送エラーと見做し、エラー終了。
	  (cond [(= 200 (lws:code-of res))
			 (let* [(sxml  (html->sxml (open-input-string (lws:content-of res))))
					(links ((sxpath (string-append "//link[@rel=\"alternate\"]"
												         "[@href]"
														 "[@type=\"application/rss+xml\" or "
														  "@type=\"text/xml\"]")) sxml))]
			   (define (link->feed link)
				 `((feed-uri . ,(car ((sxpath "@href/text()") link)))
				   
				   ,@(let [(title ((sxpath "@title/text()") link))]
					   (if (null? title)
						   '()
						   `((title . ,(car title)))))))
			 
			   (map link->feed links))]

			[else
			 (error `(transport-error . ,(lws:code-of res)))]))))


(define (feed:unsubscribe arg db-pod)
  (let [(auth (get-arg arg 'auth     check: list?  ))
		(uri  (get-arg arg 'feed-uri check: string?))]

	(account:check auth db-pod)

	(set! uri (util:clean-uri uri))

	(let [(feed-id (feed:find-feed-id-by-uri db-pod uri))]
	  (when feed-id
		
		(db-pod:run db-pod
		    (lambda (db)
			  (sqlite3:exec db "BEGIN IMMEDIATE")
		  
			  ;; 先ず subscription から削除する。
			  (sqlite3:exec db (string-append "DELETE FROM subscription"
											  " WHERE feed_id = ? AND  "
											  "       user_id = ?      ")
							feed-id
							(get-arg auth 'user-id))
			  ;; このフィードを登録してゐる者が誰も居なくなった場合は、フィー
			  ;; ド自體を削除する。
			  (when (= 0 (sqlite3:first-result
						  db (string-append "SELECT COUNT(*)    "
											"  FROM subscription"
											" WHERE feed_id = ? ")
						  feed-id))
				(sqlite3:exec db (string-append "DELETE FROM feeds "
												" WHERE feed_id = ?")
							  feed-id))

              (sqlite3:exec db "COMMIT")))))

	'ok))


(define (feed:set-rank arg db-pod)
  (let [(auth (get-arg arg 'auth     check: list?  ))
		(uri  (get-arg arg 'feed-uri check: string?))
		(rank (get-arg arg 'rank     check: (lambda (n)
										  (and (integer? n)
											   (<= 1 n)
											   (<= n 5)))))]
	(account:check auth db-pod)

	(set! uri (util:clean-uri uri))

	(let [(feed-id (feed:find-feed-id-by-uri db-pod uri))]
	  (when feed-id

		(db-pod:run db-pod
	        (lambda (db)
			  (sqlite3:exec db (string-append "UPDATE subscription   "
											  "   SET rank    = ?    "
											  " WHERE feed_id = ? AND"
											  "       user_id = ?    ")
							rank
							feed-id
							(get-arg auth 'user-id))))))
	
	'ok))
	  

(define (feed:get-info arg db-pod)
  (let [(auth (get-arg arg 'auth     check: list?  ))
		(uri  (get-arg arg 'feed-uri check: string?))]

	(account:check auth db-pod)

	(set! uri (util:clean-uri uri))

	(define feed-list (feed:list `((auth      . ,auth)
								   (need-desc . true )
                                   (feed-uri  . ,(uri->string uri)))
								 db-pod))

	(if (null? feed-list)
		(error (sprintf "You have not subscribed to `~a'."
						(uri->string uri))))

	(car feed-list)))


(define (feed:list arg db-pod)
  (let [(auth      (get-arg arg 'auth      check: list?                      ))
		(need-desc (get-arg arg 'need-desc check: true-or-false? optional: #t))
        (feed-uri  (get-arg arg 'feed-uri  check: string?        optional: #t))]

	(account:check auth db-pod)

    (define feed-uri-sql (if feed-uri
                             (sprintf "AND F.feed_uri = ~a"
                                      (util:sqlite-quote feed-uri char: #\"))
                             ""))

	(db-pod:run db-pod
	    (lambda (db)
		  (sqlite3:map-row
		      (lambda (feed-id feed-uri site-uri lastmod title description status rank)

                ;; all-items と read-items は今から別クエリで算出。
                ;; unread-items はその二つから算出。
                (match (sqlite3:first-row db (string-append
                                              "SELECT COUNT(*)        ,               "
                                              "       COUNT(R.item_id)                "
                                              "  FROM      items      AS I            "
                                              "  LEFT JOIN read_items AS R            "
                                              "         ON (R.item_id = I.item_id AND "
                                              "             R.user_id = ?            )"
                                              " WHERE I.feed_id = ?                   ")
                                          (get-arg auth 'user-id)
                                          feed-id)
                       [(all-items read-items)
                        ;; 常にこれにマッチする
                        `((feed-uri     . ,feed-uri               )
                          (site-uri     . ,site-uri               )
                          (lastmod      . ,lastmod                )
                          (title        . ,title                  )
                          ,@(if (eq? need-desc 'true)
                                `((description . ,description))
                                '())
                          (status       . ,(string->symbol status))
                          (rank         . ,rank                   )
                          (all-items    . ,all-items              )
                          (read-items   . ,read-items             )
                          (unread-items . ,(- all-items read-items)))]))
			  db
			  (string-append "SELECT F.feed_id    ,                 "
                             "       F.feed_uri   ,                 "
                             "       F.site_uri   ,                 "
							 "       F.lastmod    ,                 "
							 "       F.title      ,                 "
							 "       F.description,                 "
							 "       F.status     ,                 "
                             "       S.rank                         "
							 "  FROM subscription AS S              "
							 "  LEFT JOIN feeds AS F USING (feed_id)"
							 " WHERE S.user_id = ?                  "
                             "     " feed-uri-sql                    )
			  (get-arg auth 'user-id))))))
