;; -*- hen -*-
#|
  [ S-exp RPC ]

  * 概要

  sRPC は通信内容に S 式を用ゐる RPC プロトコルである。JSON-RPC の思想
  を繼承してゐる。

  * 方針

  sRPC-Node 自體の實裝が容易である事、及び sRPC を用ゐるアプリケーショ
  ンの實裝が容易である事を重視する。

  * ノードと動作モード

  sRPC ではサーバーとクライアントの區別は無く、サービスを提供する側と提
  供を受ける側が、どちらも等しくリクエストを送信し、應答を受信する事が
  出來る。サービスの提供側、すなはち TCP/IP を用ゐて特定のポートを
  listen したり、特定のファイルパスにて UNIX ドメインソケットを listen
  したりする側を『受動的なノード (Passive Node)』と言ひ、そのやうに開か
  れてゐるソケットへ接續する側を『能動的なノード (Active Node)』と言ふ。

  * 通信内容

    送受信される S 式には、要求 (Request)、應答 (Response)、通知
    (Notification) の三種類があり、何れも S 式中に改行コード LF を含ま
    ず末尾に改行コード LF を持つ。S 式を複數の行に跨って送信する事は出
    來ず、また一行で複數の S 式を送信する事も出來ない。

    ** 要求

       ((method . SYMBOL)
	(arg    . ANY   )
	(id     . ANY   ))

       要求は、method, arg, id の三要素を持つ alist で表される。

       *** method

           この要求のメソッド名。省略不可能。シンボルでなければならない。

       *** arg

           この要求の引數。省略可能。引數として渡す事が可能なオブジェク
           トは、シンボル、整數、實數、cons セル、リスト、文字列の何れ
           かである。また、文字列は UTF-8 でエンコードされていなければ
           ならず、利用可能なエスケープシーケンスは \\, \n 及び \" のみ
           である。

       *** id

           送信側でこの要求を識別する爲に一意に與へる値。引數として使へ
           るものであれば何でも良い。

       要求を受信したノードは、必ずしも要求を受信した順序で應答する必要
       は無い。先に送った要求に對する應答が後になって返って來る可能性が
       ある爲、id によって各要求を識別する必要がある。

    ** 應答

       ((result . ANY) (id . ANY)) もしくは ((error . ANY) (id . ANY))
       
       *** result

           成功時の應答。省略可能。error と同時に存在してはならない。

       *** error

           失敗時の應答。省略不可能。result と同時に存在してはならない。

       *** id

           要求に付いてゐた id と同じもの。

    ** 通知

       ((method . SYMBOL)
	(arg    . ANY   ))

       id が無い他は要求と同じ。要求とは異なり、通知に對しては應答が返
       されない。

|#

#|
    [實裝について]

    簡單化の爲、特に必要になるまでは Passive から Active への要求送信は
    實裝せず、また Chicken では Passive のみを、EmacsLisp では Active の
    みを實裝する。
|#

(declare (unit srpc-passive-node)
		 (uses logger)
		 (export <srpc-passive-node>
				 initialize
				 start!
				 register-method!
				 get-method))

(include "cleited-macros.scm")

(use tinyclos
     tcp
     extras
     tcp-server)

(define-class <srpc-passive-node> ()
  (port            ;; listen するポート
   server-thread   ;; listen してゐるスレッド
   max-requests    ;; 最大同時接續數
   methods         ;; メソッド名シンボル => (lambda (arg) ...)
   methods-mutex)) ;; その mutex

(define-method (initialize (this <srpc-passive-node>) args)
  (call-next-method)
  (define (init #!key port max-requests)

	(unless port
	  (set! port 9895))

	(assert (integer? port))
	(assert (or (eq? #f max-requests)
				(positive? max-requests)))

	(initialize-slots this (list 'port          port
								 'server-thread (make-thread
												   (lambda ()
													 (server-main this)))
								 'max-requests  max-requests
								 'methods       (make-hash-table eq?)
								 'methods-mutex (make-mutex))))
  (apply init args))


(define-method (start! (this <srpc-passive-node>))
  ;; 別スレッドで listen を開始し、accept ループを永久に續ける。
  (with-slots (server-thread) this

     (thread-start! server-thread)))


(define-method (register-method! (this <srpc-passive-node>)
								 (name <symbol>)
								 (impl <procedure>))
  (with-slots (methods methods-mutex) this
	 (with-mutex-locking methods-mutex
						 
        (hash-table-set! methods name impl))))


(define-method (get-method (this <srpc-passive-node>)
						   (name <symbol>))
  (with-slots (methods methods-mutex) this
     (with-mutex-locking methods-mutex

	    (hash-table-ref/default methods name #f))))


(define-method (server-main (this <srpc-passive-node>))
  (with-slots (port max-requests) this

	 (let [(listener     (tcp-listen port))
		   (service-proc (lambda ()
						   (service-main this)))]

	   (if max-requests
		   ((make-tcp-server listener service-proc max-requests))
		   ((make-tcp-server listener service-proc))))))


(define-method (process-line (this <srpc-passive-node>)
							 (line <string>))
  (bind-exit exit
     (define inbound #f)

	 (condition-case
	  (with-input-from-string line
		
		(lambda ()
		  (set! inbound (read))
		  
		  (if (char-ready?)
			  (exit `((method . %parse-error)
					  (arg    . ,(format "The line has not just a single S-exp! [~a]" line)))))))
		
	  [exn (exn syntax)
		   ;; パースエラーの場合は id が分からないので
		   ;; notification でエラー應答…
		   (exit `((method . %parse-error)
				   (arg    . ,(format
							   "I can't even parse your S-exp! [~a]: ~a"
							   line
							   ((condition-property-accessor 'exn 'message) exn)))))])

	 ;; パース成功。次は受信したメッセージの種類を判別。
	 (unless (list? inbound)
	   (exit `((method . %syntax-error)
			   (arg    . ,(format "Any messages have to be an alist! [~s]" inbound)))))
	 
	 (let ((method (assq 'method inbound))
	       (arg    (assq 'arg    inbound))
	       (id     (assq 'id     inbound)))
	    
	   (unless method
		 (exit `((method . %syntax-error)
				 (arg    . ,(format "Any inbound messeges have to have a method! [~s]" inbound)))))

	   ;; メソッドは實在するものかどうか
	   (let ((method-func (get-method this (cdr method))))
	       
		 (unless method-func
		   (exit
			(if id
				`((error . ,(format "No such method is available: [~s]" (cdr method)))
				  (id    . ,(cdr id)))
		       
				`((method . %no-such-method)
				  (arg    . ,(format "No such method is available: [~s]" (cdr method)))))))
	       
		 ;; 形式的な確認は全て通過した。次にメソッドを實行するが、實行
		 ;; 時エラーが起きたら捕捉する。
		 (condition-case
		  
		  (let ((method-result
				 (method-func (if arg
								  (cdr arg) #f))))
            #|
            (logger 'srpc-passive-node 'info
                    "sRPC method call by ~a: method ~s: arg ~s: result ~s"
                    (tcp-addresses (current-input-port))
                    (cdr method)
                    (if arg (cdr arg) #f)
                    method-result)
            |#
			;; id があれば應答を返すが、無ければ何も返さない。何も返さ
			;; ない場合は、このメソッドは #f を返す。
			(if id
			    `((result . ,method-result)
			      (id     . ,(cdr id)))
			    
			    #f))

		  [exn ()
			   ;;(raise exn) ; FIXME: delete this
			   
			   (logger 'srpc-passive-node 'error
					   "sRPC call error: method ~a:~%~a"
					   (cdr method)
					   (with-output-to-string
						 (lambda ()
						   (print-error-message exn)
						   (print-call-chain))))
			   
			   (let ((msg ((condition-property-accessor 'exn 'message) exn)))
				 (exit
				  (if id
					  `((error . ,msg)
						(id    . ,(cdr id)))
			  
					  `((method . %runtime-error)
						(arg    . ,msg)))))])))))


(define-method (service-main (this <srpc-passive-node>))
  (logger 'srpc-passive-node 'info
          "accepted connection from ~a"
          (tcp-addresses (current-input-port)))
  
  (let loop ((line (read-line)))
	(if (eof-object? line)
        ;; 切斷
        (logger 'srpc-passive-node 'info
                "connection closed by ~a"
                (tcp-addresses (current-input-port)))

        (begin
          ;; とりあえず１行讀めたのでパースしてみる
          (let ((outbound (process-line this line)))
            ;; 出力内容が決まった
            (if outbound
                (begin
                  (printf "~s~%" outbound)
                  (flush-output))))

          (loop (read-line))))))
