;; -*- hen -*-
(declare (unit uri)
		 (export <uri>
				 initialize
				 print-object
				 
				 uri:parse!
				 uri:encode
				 uri:canonicalize
				 uri:canonicalize!
				 uri:default-port
				 uri:->string
				 uri->string

				 uri:scheme-of
				 uri:userinfo-of
				 uri:host-of
				 uri:port-of
				 uri:path-of
				 uri:query-of
				 uri:fragment-of

				 uri:set-scheme!
				 uri:set-userinfo!
				 uri:set-host!
				 uri:set-path!
				 uri:set-query!
				 uri:set-fragment!

				 uri:canonical-escape))

(use tinyclos
	 regex
	 regex-case
	 hostinfo
	 (srfi 13))

(include "cleited-macros.scm")

(define-class <uri> ()
  (scheme
   userinfo
   host
   port
   path
   query
   fragment))

(define-accessors <uri> (scheme
						 userinfo
						 host
						 port
						 path
						 query
						 fragment)
  prefix: uri:)

(define uri:re-uri
  (regexp (string-append "^"
						 "("              ;; $1
						    "([^:/?#]+):" ;; $2  - scheme
						 ")?"
						 "("              ;; $3
						    "//([^/?#]*)" ;; $4  - authority
						 ")?"
						 "("              ;; $5  - path
						    "[^?#]*"
						 ")"
						 "("              ;; $7
						    "\\?([^#]*)"  ;; $8  - query
						 ")?"
						 "("              ;; $9
						    "#(.*)"       ;; $10 - fragment
						 ")?"
						 "$")))

(define uri:re-authority
  (regexp (string-append "^"
						 "("              ;; $1
						    "([^@]*)@"    ;; $2  - userinfo
						 ")?"
						 "("              ;; $3
						    "\\[([0-9a-fA-F:]+)\\]|" ;; $4 - host (ipv6)
							"([^:]+)"     ;; $5  - host (hostname or ipv4)
						 ")"
						 "("              ;; $6
						    ":([0-9]+)"   ;; $7  - port
						 ")?"
						 "$")))

(define uri:re-uric
  (regexp (string-append "("
						    "["
						       ";/?:@&=+$," ;; reserved
							   "_.!~*'()"         ;; mark
							   "A-Za-z0-9"        ;; unreserved
							   "%"                ;; %
						    "]|"
							"-|"
						    "\\[|"
							"\\]"
						 ")")))

(define-method (initialize (this <uri>) args)
  (call-next-method)

  (define (init uri #!optional base-uri)
	
	(cond ((instance-of? uri <uri>)
		   (uri:parse! this (uri:->string uri)))

		  ((string? uri)
		   (uri:parse! this uri))

		  (else
		   (error "invalid uri argument")))

	(if base-uri
		(error "FIXME: not implemented yet"))
		;;(uri:abs! this base-uri))

	this)

  (apply init args))

(define-method (print-object (this <uri>) #!optional (port (current-output-port)))
  (fprintf port "#<uri (~a)>" (uri:->string this)))

(define-method (uri:parse! (this <uri>) uri)
  (regex-case uri
     [uri:re-uri (_ _ scheme _ authority path _ query _ fragment)

				 (slot-set! this 'scheme
							(if scheme
								;; scheme はエンコードしない
								scheme
								#f))

				 (if authority
					 ;; authority らしきものが存在する。
					 (uri:parse-authority! this authority))

				 ;; path は常に存在
				 (slot-set! this 'path (uri:encode path))

				 (slot-set! this 'query
							(if query
								(uri:encode query)
								#f))

				 (slot-set! this 'fragment
							(if fragment
								(uri:encode fragment)
								#f))]
	 
	 [else (error (sprintf "failed to parse uri: ~a" uri))]))

(define-method (uri:parse-authority! (this <uri>) authority)
  (regex-case authority
     [uri:re-authority (_ _ userinfo _ host-ipv6 host-other _ port)

					   (slot-set! this 'userinfo
								  (if userinfo
									  (uri:encode userinfo)
									  #f))

					   (slot-set! this 'host
								  (uri:encode (or host-ipv6 host-other)))

					   (slot-set! this 'port
								  (if port
									  (string->number port)
									  #f))]

	 [else (error (sprintf "failed to parse authority: ~a" authority))]))

(define (hex? ch)
	(or (and (char>=? ch #\0) (char<=? ch #\9))
		(and (char>=? ch #\a) (char<=? ch #\f))
		(and (char>=? ch #\A) (char<=? ch #\F))))

(define (int->hex n)
  (assert (and (>= n 0)
			   (<= n 15)))
  (integer->char (if (< n 10)
					 (+ (char->integer #\0) n)
					 (+ (char->integer #\A) (- n 10)))))

(define (hex->int ch)
  (assert (hex? ch))
  (cond ((and (char>=? ch #\0) (char<=? ch #\9))
		 (- (char->integer ch) (char->integer #\0)))

		((and (char>=? ch #\a) (char<=? ch #\f))
		 (+ 10
			(- (char->integer ch) (char->integer #\a))))

		((and (char>=? ch #\A) (char<=? ch #\F))
		 (+ 10
			(- (char->integer ch) (char->integer #\A))))))

(define (uri:decode str)
  
  (define (decode l)
	(if (null? l)
		l
		
		(let ((x  (car l))
			  (xs (cdr l)))

		  (if (and (char=? x #\%)
				   (>=   (length xs) 2)
				   (hex? (car   xs))
				   (hex? (cadr  xs)))
			  ;; %xx の形になってゐる
			  (let ((hi-4bits (hex->int (car  xs)))
					(lo-4bits (hex->int (cadr xs))))

				(cons (integer->char (bitwise-ior (arithmetic-shift hi-4bits 4)
												  lo-4bits))
					  (decode (cddr xs))))
			  
			  ;; なってゐない
			  (cons x (decode xs))))))

  (list->string (decode (string->list str))))
   

(define (uri:encode str)
  (define (encode ch)
	(regex-case (string ch)
				
	   [uri:re-uric _ ;; uric なので無變換
					(string ch)]

	   [else ;; %xx にエンコード
		(let* ((num      (char->integer ch))
			   (hi-4bits (bitwise-and (arithmetic-shift num -4) 15)) ;; 15: 0x0F
			   (lo-4bits (bitwise-and num 15)))

		  (string #\%
				  (int->hex hi-4bits)
				  (int->hex lo-4bits)))]))

  ;; (apply string-append (map... )) をやると引數の個數制限に引っ掛る。
  (string-concatenate (map encode (string->list str))))

(define-method (uri:canonicalize (this <uri>))
  (let ((u (make <uri> this)))
	
	(uri:canonicalize! u)
	u))

(define-method (uri:canonicalize! (this <uri>))
  (with-slots (scheme userinfo host port path query fragment) this

     (if scheme
		 (begin
		   (string-downcase! scheme)

		   (if (and port
					(= (uri:default-port this) port))
			   
			   (slot-set! this 'port #f))))

	 (if host
		 (string-downcase! host)
		 ;; host が無いなら userinfo と port は消す
		 (begin
		   (slot-set! this 'userinfo #f)
		   (slot-set! this 'port     #f)))

	 ;; RFC-2616 3.2.3 節によると、空の path は "/" と等価である。
	 (if (equal? path "")
		 (slot-set! path "/"))

	 (slot-set! this 'userinfo (uri:canonical-escape userinfo))
	 (slot-set! this 'host     (uri:canonical-escape host))
	 (slot-set! this 'path     (uri:canonical-escape path))
	 (slot-set! this 'query    (uri:canonical-escape query))
	 (slot-set! this 'fragment (uri:canonical-escape fragment))))

(define (uri:canonical-escape str)
  (if str
	  (begin
		(assert (string? str))
		(uri:encode (uri:decode str)))

	  #f))

(define-method (uri:default-port (this <uri>))
  (with-slots (scheme) this

     (if scheme
		 (service-name->port scheme)
		 #f)))

(define-method (uri->string (this <uri>))
  (uri:->string this))

(define-method (uri:host-of (this <uri>))
  (with-slots (host) this
      host))

(define-method (uri:port-of (this <uri>))
  (with-slots (port) this
	  port))

(define-method (uri:->string (this <uri>))
  (with-slots (scheme userinfo host port path query fragment) this

      (with-output-to-string
		(lambda ()

		  (if scheme
			  (begin (display scheme)
					 (display #\:)))

		  (if host
			  (begin
				(display "//")

				(if userinfo
					(begin (display userinfo)
						   (display #\@)))

				(if (string-contains host ":")
					;; IPv6 アドレス
					(begin (display #\[)
						   (display host)
						   (display #\]))
					;; それ以外
					(display host))

				(if port
					(begin (display #\:)
						   (display port)))

				;; host が在るなら path は絶對パスでなければならない。も
				;; しさうでなければ生成される URI が不正なものになる。
				(if (and path
						 (not (and (> (string-length path) 0)
								   (char=? (string-ref path 0) #\/))))
					(display #\/))))

		  (if path
			  (display path))

		  (if query
			  (begin (display #\?)
					 (display query)))

		  (if fragment
			  (begin (display #\#)
					 (display fragment)))))))
